import pytest
from ops.charm import CharmBase
from ops.framework import Framework

from scenario.state import State
from ..helpers import trigger


@pytest.fixture(scope="function")
def mycharm():
    class MyCharm(CharmBase):
        def __init__(self, framework: Framework):
            super().__init__(framework)
            for evt in self.on.events().values():
                self.framework.observe(evt, self._on_event)

        def _on_event(self, event):
            pass

    return MyCharm


def test_config_get(mycharm):
    def check_cfg(charm: CharmBase):
        assert charm.config["foo"] == "bar"
        assert charm.config["baz"] == 1

    trigger(
        State(
            config={"foo": "bar", "baz": 1},
        ),
        "update_status",
        mycharm,
        meta={"name": "foo"},
        config={"options": {"foo": {"type": "string"}, "baz": {"type": "int"}}},
        post_event=check_cfg,
    )


def test_config_get_default_from_meta(mycharm):
    def check_cfg(charm: CharmBase):
        assert charm.config["foo"] == "bar"
        assert charm.config["baz"] == 2
        assert charm.config["qux"] is False

    trigger(
        State(
            config={"foo": "bar"},
        ),
        "update_status",
        mycharm,
        meta={"name": "foo"},
        config={
            "options": {
                "foo": {"type": "string"},
                "baz": {"type": "int", "default": 2},
                "qux": {"type": "boolean", "default": False},
            },
        },
        post_event=check_cfg,
    )


@pytest.mark.parametrize(
    "cfg_in",
    (
        {"foo": "bar"},
        {"baz": 4, "foo": "bar"},
        {"baz": 4, "foo": "bar", "qux": True},
    ),
)
def test_config_in_not_mutated(mycharm, cfg_in):
    class MyCharm(CharmBase):
        def __init__(self, framework: Framework):
            super().__init__(framework)
            for evt in self.on.events().values():
                self.framework.observe(evt, self._on_event)

        def _on_event(self, event):
            # access the config to trigger a config-get
            foo_cfg = self.config["foo"]  # noqa: F841
            baz_cfg = self.config["baz"]  # noqa: F841
            qux_cfg = self.config["qux"]  # noqa: F841

    state_out = trigger(
        State(
            config=cfg_in,
        ),
        "update_status",
        MyCharm,
        meta={"name": "foo"},
        config={
            "options": {
                "foo": {"type": "string"},
                "baz": {"type": "int", "default": 2},
                "qux": {"type": "boolean", "default": False},
            },
        },
    )
    # check config was not mutated by scenario
    assert state_out.config == cfg_in



#!/usr/bin/env python3
# Copyright 2024 Canonical Ltd.
# See LICENSE file for licensing details.

"""Exceptions raised by the framework.

Note that these exceptions are not meant to be caught by charm authors. They are
used by the framework to signal errors or inconsistencies in the charm tests
themselves.
"""


class ContextSetupError(RuntimeError):
    """Raised by Context when setup fails."""


class AlreadyEmittedError(RuntimeError):
    """Raised when ``run()`` is called more than once."""


class ScenarioRuntimeError(RuntimeError):
    """Base class for exceptions raised by the runtime module."""


class UncaughtCharmError(ScenarioRuntimeError):
    """Error raised if the charm raises while handling the event being dispatched."""


class InconsistentScenarioError(ScenarioRuntimeError):
    """Error raised when the combination of state and event is inconsistent."""


class StateValidationError(RuntimeError):
    """Raised when individual parts of the State are inconsistent."""

    # as opposed to InconsistentScenario error where the **combination** of
    # several parts of the State are.


class MetadataNotFoundError(RuntimeError):
    """Raised when a metadata file can't be found in the provided charm root."""


class ActionMissingFromContextError(Exception):
    """Raised when the user attempts to invoke action hook tools outside an action context."""

    # This is not an ops error: in ops, you'd have to go exceptionally out of
    # your way to trigger this flow.


class NoObserverError(RuntimeError):
    """Error raised when the event being dispatched has no registered observers."""


class BadOwnerPath(RuntimeError):
    """Error raised when the owner path does not lead to a valid ObjectEvents instance."""



#!/usr/bin/env python3
# Copyright 2023 Canonical Ltd.
# See LICENSE file for licensing details.

"""Test Context

The test `Context` object provides the context of the wider Juju system that the
specific `State` exists in, and the events that can be executed on that `State`.
"""

from __future__ import annotations

import functools
import tempfile
from contextlib import contextmanager
from pathlib import Path
from typing import (
    Generic,
    TYPE_CHECKING,
    Any,
    Callable,
    Mapping,
)

import ops
from ops._private.harness import ActionFailed

from .errors import (
    AlreadyEmittedError,
    ContextSetupError,
    MetadataNotFoundError,
)
from .logger import logger as scenario_logger
from .state import (
    CharmType,
    CheckInfo,
    Container,
    Notice,
    Secret,
    Storage,
    _Action,
    _CharmSpec,
    _Event,
)
from ._runtime import Runtime

if TYPE_CHECKING:  # pragma: no cover
    from ops._private.harness import ExecArgs
    from ._ops_main_mock import Ops
    from .state import (
        AnyJson,
        CharmType,
        JujuLogLine,
        RelationBase,
        State,
        _EntityStatus,
    )

logger = scenario_logger.getChild("runtime")

_DEFAULT_JUJU_VERSION = "3.5"


class Manager(Generic[CharmType]):
    """Context manager to offer test code some runtime charm object introspection.

    This class should not be instantiated directly: use a :class:`Context`
    in a ``with`` statement instead, for example::

        ctx = Context(MyCharm)
        with ctx(ctx.on.start(), State()) as manager:
            manager.charm.setup()
            manager.run()
    """

    def __init__(
        self,
        ctx: Context[CharmType],
        arg: _Event,
        state_in: State,
    ):
        self._ctx = ctx
        self._arg = arg
        self._state_in = state_in

        self._emitted: bool = False

        self.ops: Ops[CharmType] | None = None

    @property
    def charm(self) -> CharmType:
        """The charm object instantiated by ops to handle the event.

        The charm is only available during the context manager scope.
        """
        if self.ops is None or self.ops.charm is None:
            raise RuntimeError(
                "you should __enter__ this context manager before accessing this",
            )
        return self.ops.charm

    @property
    def _runner(self):
        return self._ctx._run  # noqa

    def __enter__(self):
        self._wrapped_ctx = wrapped_ctx = self._runner(self._arg, self._state_in)
        self.ops = wrapped_ctx.__enter__()
        return self

    def run(self) -> State:
        """Emit the event and proceed with charm execution.

        This can only be done once.
        """
        if self._emitted:
            raise AlreadyEmittedError("Can only run once.")
        if not self.ops:
            raise RuntimeError(
                "you should __enter__ this context manager before running it",
            )
        self._emitted = True
        self.ops.run()

        # wrap up Runtime.exec() so that we can gather the output state
        self._wrapped_ctx.__exit__(None, None, None)

        assert self._ctx._output_state is not None
        return self._ctx._output_state

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any):  # noqa: U100
        if not self._emitted:
            logger.debug(
                "user didn't emit the event within the context manager scope. "
                "Doing so implicitly upon exit...",
            )
            self.run()


def _copy_doc(original_func: Callable[..., Any]):
    """Copy the docstring from `original_func` to the wrapped function."""

    def decorator(wrapper_func: Callable[..., Any]):
        @functools.wraps(wrapper_func)
        def wrapped(*args: Any, **kwargs: Any):
            return wrapper_func(*args, **kwargs)

        wrapped.__doc__ = original_func.__doc__
        return wrapped

    return decorator


class CharmEvents:
    """Events generated by Juju or ops pertaining to the application lifecycle.

    The events listed as attributes of this class should be accessed via the
    :attr:`Context.on` attribute. For example::

        ctx.run(ctx.on.config_changed(), state)

    This behaves similarly to the :class:`ops.CharmEvents` class but is much
    simpler as there are no dynamically named attributes, and no ``__getattr__``
    version to get events. In addition, all of the attributes are methods,
    which are used to connect the event to the specific object that they relate
    to (or, for simpler events like "start" or "stop", take no arguments).
    """

    @staticmethod
    @_copy_doc(ops.InstallEvent)
    def install():
        return _Event("install")

    @staticmethod
    @_copy_doc(ops.StartEvent)
    def start():
        return _Event("start")

    @staticmethod
    @_copy_doc(ops.StopEvent)
    def stop():
        return _Event("stop")

    @staticmethod
    @_copy_doc(ops.RemoveEvent)
    def remove():
        return _Event("remove")

    @staticmethod
    @_copy_doc(ops.UpdateStatusEvent)
    def update_status():
        return _Event("update_status")

    @staticmethod
    @_copy_doc(ops.ConfigChangedEvent)
    def config_changed():
        return _Event("config_changed")

    @staticmethod
    @_copy_doc(ops.UpgradeCharmEvent)
    def upgrade_charm():
        return _Event("upgrade_charm")

    @staticmethod
    @_copy_doc(ops.PreSeriesUpgradeEvent)
    def pre_series_upgrade():
        return _Event("pre_series_upgrade")

    @staticmethod
    @_copy_doc(ops.PostSeriesUpgradeEvent)
    def post_series_upgrade():
        return _Event("post_series_upgrade")

    @staticmethod
    @_copy_doc(ops.LeaderElectedEvent)
    def leader_elected():
        return _Event("leader_elected")

    @staticmethod
    @_copy_doc(ops.SecretChangedEvent)
    def secret_changed(secret: Secret):
        if secret.owner:
            raise ValueError(
                "This unit will never receive secret-changed for a secret it owns.",
            )
        return _Event("secret_changed", secret=secret)

    @staticmethod
    @_copy_doc(ops.SecretExpiredEvent)
    def secret_expired(secret: Secret, *, revision: int):
        if not secret.owner:
            raise ValueError(
                "This unit will never receive secret-expire for a secret it does not own.",
            )
        return _Event("secret_expired", secret=secret, secret_revision=revision)

    @staticmethod
    @_copy_doc(ops.SecretRotateEvent)
    def secret_rotate(secret: Secret):
        if not secret.owner:
            raise ValueError(
                "This unit will never receive secret-rotate for a secret it does not own.",
            )
        return _Event("secret_rotate", secret=secret)

    @staticmethod
    @_copy_doc(ops.SecretRemoveEvent)
    def secret_remove(secret: Secret, *, revision: int):
        if not secret.owner:
            raise ValueError(
                "This unit will never receive secret-removed for a secret it does not own.",
            )
        return _Event("secret_remove", secret=secret, secret_revision=revision)

    @staticmethod
    def collect_app_status():
        """Event triggered at the end of every hook to collect app statuses for evaluation"""
        return _Event("collect_app_status")

    @staticmethod
    def collect_unit_status():
        """Event triggered at the end of every hook to collect unit statuses for evaluation"""
        return _Event("collect_unit_status")

    @staticmethod
    @_copy_doc(ops.RelationCreatedEvent)
    def relation_created(relation: RelationBase):
        return _Event(f"{relation.endpoint}_relation_created", relation=relation)

    @staticmethod
    @_copy_doc(ops.RelationJoinedEvent)
    def relation_joined(relation: RelationBase, *, remote_unit: int | None = None):
        return _Event(
            f"{relation.endpoint}_relation_joined",
            relation=relation,
            relation_remote_unit_id=remote_unit,
        )

    @staticmethod
    @_copy_doc(ops.RelationChangedEvent)
    def relation_changed(
        relation: RelationBase,
        *,
        remote_unit: int | None = None,
    ):
        return _Event(
            f"{relation.endpoint}_relation_changed",
            relation=relation,
            relation_remote_unit_id=remote_unit,
        )

    @staticmethod
    @_copy_doc(ops.RelationDepartedEvent)
    def relation_departed(
        relation: RelationBase,
        *,
        remote_unit: int | None = None,
        departing_unit: int | None = None,
    ):
        return _Event(
            f"{relation.endpoint}_relation_departed",
            relation=relation,
            relation_remote_unit_id=remote_unit,
            relation_departed_unit_id=departing_unit,
        )

    @staticmethod
    @_copy_doc(ops.RelationBrokenEvent)
    def relation_broken(relation: RelationBase):
        return _Event(f"{relation.endpoint}_relation_broken", relation=relation)

    @staticmethod
    @_copy_doc(ops.StorageAttachedEvent)
    def storage_attached(storage: Storage):
        return _Event(f"{storage.name}_storage_attached", storage=storage)

    @staticmethod
    @_copy_doc(ops.StorageDetachingEvent)
    def storage_detaching(storage: Storage):
        return _Event(f"{storage.name}_storage_detaching", storage=storage)

    @staticmethod
    @_copy_doc(ops.PebbleReadyEvent)
    def pebble_ready(container: Container):
        return _Event(f"{container.name}_pebble_ready", container=container)

    @staticmethod
    @_copy_doc(ops.PebbleCustomNoticeEvent)
    def pebble_custom_notice(container: Container, notice: Notice):
        return _Event(
            f"{container.name}_pebble_custom_notice",
            container=container,
            notice=notice,
        )

    @staticmethod
    @_copy_doc(ops.PebbleCheckFailedEvent)
    def pebble_check_failed(container: Container, info: CheckInfo):
        return _Event(
            f"{container.name}_pebble_check_failed",
            container=container,
            check_info=info,
        )

    @staticmethod
    @_copy_doc(ops.PebbleCheckRecoveredEvent)
    def pebble_check_recovered(container: Container, info: CheckInfo):
        return _Event(
            f"{container.name}_pebble_check_recovered",
            container=container,
            check_info=info,
        )

    @staticmethod
    @_copy_doc(ops.ActionEvent)
    def action(
        name: str,
        params: Mapping[str, AnyJson] | None = None,
        id: str | None = None,
    ):
        kwargs: dict[str, Any] = {}
        if params:
            kwargs["params"] = params
        if id:
            kwargs["id"] = id
        return _Event(f"{name}_action", action=_Action(name, **kwargs))


class Context(Generic[CharmType]):
    """Represents a simulated charm's execution context.

    The main entry point to running a test. It contains:

    - the charm source code being executed
    - the metadata files associated with it
    - a charm project repository root
    - the Juju version to be simulated

    After you have instantiated ``Context``, typically you will call :meth:`run()` to execute the
    charm once, write any assertions you like on the output state returned by the call, write any
    assertions you like on the ``Context`` attributes, then discard the ``Context``.

    Each ``Context`` instance is in principle designed to be single-use:
    ``Context`` is not cleaned up automatically between charm runs.

    Any side effects generated by executing the charm, that are not rightful part of the
    ``State``, are in fact stored in the ``Context``:

    - :attr:`juju_log`
    - :attr:`app_status_history`
    - :attr:`unit_status_history`
    - :attr:`workload_version_history`
    - :attr:`removed_secret_revisions`
    - :attr:`requested_storages`
    - :attr:`emitted_events`
    - :attr:`action_logs`
    - :attr:`action_results`

    This allows you to write assertions not only on the output state, but also, to some
    extent, on the path the charm took to get there.

    A typical test will look like::

        from charm import MyCharm, MyCustomEvent  # noqa

        def test_foo():
            # Arrange: set the context up
            ctx = Context(MyCharm)
            # Act: prepare the state and emit an event
            state_out = ctx.run(ctx.on.update_status(), State())
            # Assert: verify the output state is what you think it should be
            assert state_out.unit_status == ActiveStatus('foobar')
            # Assert: verify the Context contains what you think it should
            assert len(c.emitted_events) == 4
            assert isinstance(c.emitted_events[3], MyCustomEvent)

    If you need access to the charm object that will handle the event, use the
    class in a ``with`` statement, like::

        def test_foo():
            ctx = Context(MyCharm)
            with ctx(ctx.on.start(), State()) as manager:
                manager.charm._some_private_setup()
                manager.run()
    """

    juju_log: list[JujuLogLine]
    """A record of what the charm has sent to juju-log"""
    app_status_history: list[_EntityStatus]
    """A record of the app statuses the charm has set"""
    unit_status_history: list[_EntityStatus]
    """A record of the unit statuses the charm has set"""
    workload_version_history: list[str]
    """A record of the workload versions the charm has set"""
    removed_secret_revisions: list[int]
    """A record of the secret revisions the charm has removed"""
    emitted_events: list[ops.EventBase]
    """A record of the events (including custom) that the charm has processed"""
    requested_storages: dict[str, int]
    """A record of the storages the charm has requested"""
    action_logs: list[str]
    """The logs associated with the action output, set by the charm with :meth:`ops.ActionEvent.log`

    This will be empty when handling a non-action event.
    """
    action_results: dict[str, Any] | None
    """A key-value mapping assigned by the charm as a result of the action.

    This will be ``None`` if the charm never calls :meth:`ops.ActionEvent.set_results`
    """
    on: CharmEvents
    """The events that this charm can respond to.

    Use this when calling :meth:`run` to specify the event to emit.
    """

    def __init__(
        self,
        charm_type: type[CharmType],
        meta: dict[str, Any] | None = None,
        *,
        actions: dict[str, Any] | None = None,
        config: dict[str, Any] | None = None,
        charm_root: str | Path | None = None,
        juju_version: str = _DEFAULT_JUJU_VERSION,
        capture_deferred_events: bool = False,
        capture_framework_events: bool = False,
        app_name: str | None = None,
        unit_id: int | None = 0,
        app_trusted: bool = False,
    ):
        """Represents a simulated charm's execution context.

        If the charm, say, expects a ``./src/foo/bar.yaml`` file present relative to the
        execution cwd, you need to use the ``charm_root`` argument. For example::

            import tempfile
            virtual_root = tempfile.TemporaryDirectory()
            local_path = Path(local_path.name)
            (local_path / 'foo').mkdir()
            (local_path / 'foo' / 'bar.yaml').write_text('foo: bar')
            Context(... charm_root=virtual_root).run(...)

        :arg charm_type: the :class:`ops.CharmBase` subclass to handle the event.
        :arg meta: charm metadata to use. Needs to be a valid metadata.yaml format (as a dict).
            If none is provided, we will search for a ``metadata.yaml`` file in the charm root.
        :arg actions: charm actions to use. Needs to be a valid actions.yaml format (as a dict).
            If none is provided, we will search for a ``actions.yaml`` file in the charm root.
        :arg config: charm config to use. Needs to be a valid config.yaml format (as a dict).
            If none is provided, we will search for a ``config.yaml`` file in the charm root.
        :arg juju_version: Juju agent version to simulate.
        :arg app_name: App name that this charm is deployed as. Defaults to the charm name as
            defined in the metadata.
        :arg unit_id: Unit ID that this charm is deployed as.
        :arg app_trusted: whether the charm has Juju trust (deployed with ``--trust`` or added with
            ``juju trust``).
        :arg charm_root: virtual charm filesystem root the charm will be executed with.
        """

        if not any((meta, actions, config)):
            logger.debug("Autoloading charmspec...")
            try:
                spec: _CharmSpec[CharmType] = _CharmSpec.autoload(charm_type)
            except MetadataNotFoundError as e:
                raise ContextSetupError(
                    f"Cannot setup scenario with `charm_type`={charm_type}. "
                    f"Did you forget to pass `meta` to this Context?",
                ) from e

        else:
            if not meta:
                meta = {"name": str(charm_type.__name__)}
            spec = _CharmSpec(
                charm_type=charm_type,
                meta=meta,
                actions=actions,
                config=config,
            )

        self.charm_spec = spec
        self.charm_root = charm_root
        self.juju_version = juju_version
        if juju_version.split(".")[0] == "2":
            logger.warning(
                "Juju 2.x is closed and unsupported. You may encounter inconsistencies.",
            )

        self._app_name = app_name
        self._unit_id = unit_id
        self.app_trusted = app_trusted
        self._tmp = tempfile.TemporaryDirectory()

        # config for what events to be captured in emitted_events.
        self.capture_deferred_events = capture_deferred_events
        self.capture_framework_events = capture_framework_events

        # streaming side effects from running an event
        self.juju_log: list[JujuLogLine] = []
        self.app_status_history: list[_EntityStatus] = []
        self.unit_status_history: list[_EntityStatus] = []
        self.exec_history: dict[str, list[ExecArgs]] = {}
        self.workload_version_history: list[str] = []
        self.removed_secret_revisions: list[int] = []
        self.emitted_events: list[ops.EventBase] = []
        self.requested_storages: dict[str, int] = {}

        # set by Runtime.exec() in self._run()
        self._output_state: State | None = None

        # operations (and embedded tasks) from running actions
        self.action_logs: list[str] = []
        self.action_results: dict[str, Any] | None = None
        self._action_failure_message: str | None = None

        self.on = CharmEvents()

    def _set_output_state(self, output_state: State):
        """Hook for Runtime to set the output state."""
        self._output_state = output_state

    def _get_container_root(self, container_name: str):
        """Get the path to a tempdir where this container's simulated root will live."""
        return Path(self._tmp.name) / "containers" / container_name

    def _get_storage_root(self, name: str, index: int) -> Path:
        """Get the path to a tempdir where this storage's simulated root will live."""
        storage_root = Path(self._tmp.name) / "storages" / f"{name}-{index}"
        # in the case of _get_container_root, _MockPebbleClient will ensure the dir exists.
        storage_root.mkdir(parents=True, exist_ok=True)
        return storage_root

    def _record_status(self, state: State, is_app: bool):
        """Record the previous status before a status change."""
        if is_app:
            self.app_status_history.append(state.app_status)
        else:
            self.unit_status_history.append(state.unit_status)

    def __call__(self, event: _Event, state: State) -> Manager[CharmType]:
        """Context manager to introspect live charm object before and after the event is emitted.

        Usage::

            ctx = Context(MyCharm)
            with ctx(ctx.on.start(), State()) as manager:
                manager.charm._some_private_setup()
                manager.run()  # this will fire the event
                assert manager.charm._some_private_attribute == "bar"  # noqa

        Args:
            event: the event that the charm will respond to.
            state: the :class:`State` instance to use when handling the event.
        """
        return Manager(self, event, state)

    def run_action(self, action: str, state: State):
        """Use `run()` instead.

        :private:
        """
        raise AttributeError(
            f"call with `ctx.run`, like `ctx.run(ctx.on.action({action!r})` "
            "and find the results in `ctx.action_results`",
        )

    def run(self, event: _Event, state: State) -> State:
        """Trigger a charm execution with an event and a State.

        Calling this function will call ``ops.main`` and set up the context according to the
        specified :class:`State`, then emit the event on the charm.

        :arg event: the event that the charm will respond to. Use the :attr:`on` attribute to
            specify the event; for example: ``ctx.on.start()``.
        :arg state: the :class:`State` instance to use as data source for the hook tool calls that
            the charm will invoke when handling the event.
        """
        # Help people transition from Scenario 6:
        if isinstance(event, str):
            event = event.replace("-", "_")  # type: ignore
            if event in (
                "install",
                "start",
                "stop",
                "remove",
                "update_status",
                "config_changed",
                "upgrade_charm",
                "pre_series_upgrade",
                "post_series_upgrade",
                "leader_elected",
                "collect_app_status",
                "collect_unit_status",
            ):  # type: ignore
                suggested = f"{event}()"
            elif event in ("secret_changed", "secret_rotate"):  # type: ignore
                suggested = f"{event}(my_secret)"
            elif event in ("secret_expired", "secret_remove"):  # type: ignore
                suggested = f"{event}(my_secret, revision=1)"
            elif event in (
                "relation_created",
                "relation_joined",
                "relation_changed",
                "relation_departed",
                "relation_broken",
            ):  # type: ignore
                suggested = f"{event}(my_relation)"
            elif event in ("storage_attached", "storage_detaching"):  # type: ignore
                suggested = f"{event}(my_storage)"
            elif event == "pebble_ready":
                suggested = f"{event}(my_container)"
            elif event == "pebble_custom_notice":
                suggested = f"{event}(my_container, my_notice)"
            else:
                suggested = "event()"
            raise TypeError(
                f"call with an event from `ctx.on`, like `ctx.on.{suggested}`",
            )
        if callable(event):
            raise TypeError(
                "You should call the event method. Did you forget to add parentheses?",
            )

        if event.action:
            # Reset the logs, failure status, and results, in case the context
            # is reused.
            self.action_logs.clear()
            if self.action_results is not None:
                self.action_results.clear()
            self._action_failure_message = None
        with self._run(event=event, state=state) as ops:
            ops.run()
        # We know that the output state will have been set by this point,
        # so let the type checkers know that too.
        assert self._output_state is not None
        if event.action:
            if self._action_failure_message is not None:
                raise ActionFailed(
                    self._action_failure_message,
                    state=self._output_state,  # type: ignore
                )
        return self._output_state

    @contextmanager
    def _run(self, event: _Event, state: State):
        runtime = Runtime(
            charm_spec=self.charm_spec,
            juju_version=self.juju_version,
            charm_root=self.charm_root,
            app_name=self._app_name,
            unit_id=self._unit_id,
        )
        with runtime.exec(
            state=state,
            event=event,
            context=self,  # type: ignore
        ) as ops:
            yield ops



import dataclasses

import pytest
from ops.charm import CharmBase

from scenario._consistency_checker import check_consistency
from scenario.context import Context
from scenario.errors import InconsistentScenarioError
from scenario.state import (
    _RELATION_EVENTS_SUFFIX,
    CheckInfo,
    CloudCredential,
    CloudSpec,
    Container,
    Model,
    Network,
    Notice,
    PeerRelation,
    Relation,
    Resource,
    Secret,
    State,
    Storage,
    StoredState,
    SubordinateRelation,
    _CharmSpec,
    _Event,
)


class MyCharm(CharmBase):
    pass


def assert_inconsistent(
    state: "State",
    event: "_Event",
    charm_spec: "_CharmSpec",
    juju_version="3.0",
):
    with pytest.raises(InconsistentScenarioError):
        check_consistency(state, event, charm_spec, juju_version)


def assert_consistent(
    state: "State",
    event: "_Event",
    charm_spec: "_CharmSpec",
    juju_version="3.0",
):
    check_consistency(state, event, charm_spec, juju_version)


def test_base():
    state = State()
    event = _Event("update_status")
    spec = _CharmSpec(MyCharm, {})
    assert_consistent(state, event, spec)


def test_workload_event_without_container():
    assert_inconsistent(
        State(),
        _Event("foo-pebble-ready", container=Container("foo")),
        _CharmSpec(MyCharm, {}),
    )
    assert_consistent(
        State(containers={Container("foo")}),
        _Event("foo-pebble-ready", container=Container("foo")),
        _CharmSpec(MyCharm, {"containers": {"foo": {}}}),
    )
    assert_inconsistent(
        State(),
        _Event("foo-pebble-custom-notice", container=Container("foo")),
        _CharmSpec(MyCharm, {}),
    )
    notice = Notice("example.com/foo")
    assert_consistent(
        State(containers={Container("foo", notices=[notice])}),
        _Event("foo-pebble-custom-notice", container=Container("foo"), notice=notice),
        _CharmSpec(MyCharm, {"containers": {"foo": {}}}),
    )
    assert_inconsistent(
        State(containers={Container("foo")}),
        _Event("foo-pebble-custom-notice", container=Container("foo"), notice=notice),
        _CharmSpec(MyCharm, {"containers": {"foo": {}}}),
    )
    check = CheckInfo("http-check")
    assert_consistent(
        State(containers={Container("foo", check_infos={check})}),
        _Event("foo-pebble-check-failed", container=Container("foo"), check_info=check),
        _CharmSpec(MyCharm, {"containers": {"foo": {}}}),
    )
    assert_inconsistent(
        State(containers={Container("foo")}),
        _Event("foo-pebble-check-failed", container=Container("foo"), check_info=check),
        _CharmSpec(MyCharm, {"containers": {"foo": {}}}),
    )
    assert_consistent(
        State(containers={Container("foo", check_infos={check})}),
        _Event(
            "foo-pebble-check-recovered", container=Container("foo"), check_info=check
        ),
        _CharmSpec(MyCharm, {"containers": {"foo": {}}}),
    )
    assert_inconsistent(
        State(containers={Container("foo")}),
        _Event(
            "foo-pebble-check-recovered", container=Container("foo"), check_info=check
        ),
        _CharmSpec(MyCharm, {"containers": {"foo": {}}}),
    )
    # Ensure the check is in the correct container.
    assert_inconsistent(
        State(containers={Container("foo", check_infos={check}), Container("bar")}),
        _Event(
            "foo-pebble-check-recovered", container=Container("bar"), check_info=check
        ),
        _CharmSpec(MyCharm, {"containers": {"foo": {}, "bar": {}}}),
    )
    assert_inconsistent(
        State(containers={Container("foo", check_infos={check}), Container("bar")}),
        _Event(
            "bar-pebble-check-recovered", container=Container("bar"), check_info=check
        ),
        _CharmSpec(MyCharm, {"containers": {"foo": {}, "bar": {}}}),
    )


def test_container_meta_mismatch():
    assert_inconsistent(
        State(containers={Container("bar")}),
        _Event("foo"),
        _CharmSpec(MyCharm, {"containers": {"baz": {}}}),
    )
    assert_consistent(
        State(containers={Container("bar")}),
        _Event("foo"),
        _CharmSpec(MyCharm, {"containers": {"bar": {}}}),
    )


def test_container_in_state_but_no_container_in_meta():
    assert_inconsistent(
        State(containers={Container("bar")}),
        _Event("foo"),
        _CharmSpec(MyCharm, {}),
    )
    assert_consistent(
        State(containers={Container("bar")}),
        _Event("foo"),
        _CharmSpec(MyCharm, {"containers": {"bar": {}}}),
    )


def test_container_not_in_state():
    container = Container("bar")
    assert_inconsistent(
        State(),
        _Event("bar_pebble_ready", container=container),
        _CharmSpec(MyCharm, {"containers": {"bar": {}}}),
    )
    assert_consistent(
        State(containers={container}),
        _Event("bar_pebble_ready", container=container),
        _CharmSpec(MyCharm, {"containers": {"bar": {}}}),
    )


def test_evt_bad_container_name():
    assert_inconsistent(
        State(),
        _Event("foo-pebble-ready", container=Container("bar")),
        _CharmSpec(MyCharm, {}),
    )
    assert_consistent(
        State(containers={Container("bar")}),
        _Event("bar-pebble-ready", container=Container("bar")),
        _CharmSpec(MyCharm, {"containers": {"bar": {}}}),
    )


@pytest.mark.parametrize("suffix", _RELATION_EVENTS_SUFFIX)
def test_evt_bad_relation_name(suffix):
    assert_inconsistent(
        State(),
        _Event(f"foo{suffix}", relation=Relation("bar")),
        _CharmSpec(MyCharm, {"requires": {"foo": {"interface": "xxx"}}}),
    )
    relation = Relation("bar")
    assert_consistent(
        State(relations={relation}),
        _Event(f"bar{suffix}", relation=relation),
        _CharmSpec(MyCharm, {"requires": {"bar": {"interface": "xxx"}}}),
    )


@pytest.mark.parametrize("suffix", _RELATION_EVENTS_SUFFIX)
def test_evt_no_relation(suffix):
    assert_inconsistent(State(), _Event(f"foo{suffix}"), _CharmSpec(MyCharm, {}))
    relation = Relation("bar")
    assert_consistent(
        State(relations={relation}),
        _Event(f"bar{suffix}", relation=relation),
        _CharmSpec(MyCharm, {"requires": {"bar": {"interface": "xxx"}}}),
    )


def test_config_key_missing_from_meta():
    assert_inconsistent(
        State(config={"foo": True}),
        _Event("bar"),
        _CharmSpec(MyCharm, {}),
    )
    assert_consistent(
        State(config={"foo": True}),
        _Event("bar"),
        _CharmSpec(MyCharm, {}, config={"options": {"foo": {"type": "boolean"}}}),
    )


def test_bad_config_option_type():
    assert_inconsistent(
        State(config={"foo": True}),
        _Event("bar"),
        _CharmSpec(MyCharm, {}, config={"options": {"foo": {"type": "string"}}}),
    )
    assert_inconsistent(
        State(config={"foo": True}),
        _Event("bar"),
        _CharmSpec(MyCharm, {}, config={"options": {"foo": {}}}),
    )
    assert_consistent(
        State(config={"foo": True}),
        _Event("bar"),
        _CharmSpec(MyCharm, {}, config={"options": {"foo": {"type": "boolean"}}}),
    )


@pytest.mark.parametrize(
    "config_type",
    (
        ("string", "foo", 1),
        ("int", 1, "1"),
        ("float", 1.0, 1),
        ("boolean", False, "foo"),
    ),
)
def test_config_types(config_type):
    type_name, valid_value, invalid_value = config_type
    assert_consistent(
        State(config={"foo": valid_value}),
        _Event("bar"),
        _CharmSpec(MyCharm, {}, config={"options": {"foo": {"type": type_name}}}),
    )
    assert_inconsistent(
        State(config={"foo": invalid_value}),
        _Event("bar"),
        _CharmSpec(MyCharm, {}, config={"options": {"foo": {"type": type_name}}}),
    )


@pytest.mark.parametrize("juju_version", ("3.4", "3.5", "4.0"))
def test_config_secret(juju_version):
    assert_consistent(
        State(config={"foo": "secret:co28kefmp25c77utl3n0"}),
        _Event("bar"),
        _CharmSpec(MyCharm, {}, config={"options": {"foo": {"type": "secret"}}}),
        juju_version=juju_version,
    )
    assert_inconsistent(
        State(config={"foo": 1}),
        _Event("bar"),
        _CharmSpec(MyCharm, {}, config={"options": {"foo": {"type": "secret"}}}),
    )
    assert_inconsistent(
        State(config={"foo": "co28kefmp25c77utl3n0"}),
        _Event("bar"),
        _CharmSpec(MyCharm, {}, config={"options": {"foo": {"type": "secret"}}}),
    )
    assert_inconsistent(
        State(config={"foo": "secret:secret"}),
        _Event("bar"),
        _CharmSpec(MyCharm, {}, config={"options": {"foo": {"type": "secret"}}}),
    )
    assert_inconsistent(
        State(config={"foo": "secret:co28kefmp25c77utl3n!"}),
        _Event("bar"),
        _CharmSpec(MyCharm, {}, config={"options": {"foo": {"type": "secret"}}}),
    )


@pytest.mark.parametrize("juju_version", ("2.9", "3.3"))
def test_config_secret_old_juju(juju_version):
    assert_inconsistent(
        State(config={"foo": "secret:co28kefmp25c77utl3n0"}),
        _Event("bar"),
        _CharmSpec(MyCharm, {}, config={"options": {"foo": {"type": "secret"}}}),
        juju_version=juju_version,
    )


@pytest.mark.skip(
    "The right exception is raised but pytest.raises doesn't catch it - figure this out!"
)
@pytest.mark.parametrize("bad_v", ("1.0", "0", "1.2", "2.35.42", "2.99.99", "2.99"))
def test_secrets_jujuv_bad(bad_v):
    secret = Secret({"a": "b"})
    assert_inconsistent(
        State(secrets={secret}),
        _Event("bar"),
        _CharmSpec(MyCharm, {}),
        bad_v,
    )
    assert_inconsistent(
        State(secrets={secret}),
        _Event("secret_changed", secret=secret),
        _CharmSpec(MyCharm, {}),
        bad_v,
    )

    assert_inconsistent(
        State(),
        _Event("secret_changed", secret=secret),
        _CharmSpec(MyCharm, {}),
        bad_v,
    )


@pytest.mark.parametrize("good_v", ("3.0", "3.1", "3", "3.33", "4", "100"))
def test_secrets_jujuv_good(good_v):
    assert_consistent(
        State(secrets={Secret({"a": "b"})}),
        _Event("bar"),
        _CharmSpec(MyCharm, {}),
        good_v,
    )


def test_secret_not_in_state():
    secret = Secret({"a": "b"})
    assert_inconsistent(
        State(),
        _Event("secret_changed", secret=secret),
        _CharmSpec(MyCharm, {}),
    )
    assert_consistent(
        State(secrets={secret}),
        _Event("secret_changed", secret=secret),
        _CharmSpec(MyCharm, {}),
    )


def test_peer_relation_consistency():
    assert_inconsistent(
        State(relations={Relation("foo")}),
        _Event("bar"),
        _CharmSpec(MyCharm, {"peers": {"foo": {"interface": "bar"}}}),
    )
    assert_consistent(
        State(relations={PeerRelation("foo")}),
        _Event("bar"),
        _CharmSpec(MyCharm, {"peers": {"foo": {"interface": "bar"}}}),
    )


def test_duplicate_endpoints_inconsistent():
    assert_inconsistent(
        State(),
        _Event("bar"),
        _CharmSpec(
            MyCharm,
            {
                "requires": {"foo": {"interface": "bar"}},
                "provides": {"foo": {"interface": "baz"}},
            },
        ),
    )


def test_sub_relation_consistency():
    assert_inconsistent(
        State(relations={Relation("foo")}),
        _Event("bar"),
        _CharmSpec(
            MyCharm,
            {"requires": {"foo": {"interface": "bar", "scope": "container"}}},
        ),
    )

    assert_consistent(
        State(relations={SubordinateRelation("foo")}),
        _Event("bar"),
        _CharmSpec(
            MyCharm,
            {"requires": {"foo": {"interface": "bar", "scope": "container"}}},
        ),
    )


def test_relation_sub_inconsistent():
    assert_inconsistent(
        State(relations={SubordinateRelation("foo")}),
        _Event("bar"),
        _CharmSpec(MyCharm, {"requires": {"foo": {"interface": "bar"}}}),
    )


def test_relation_not_in_state():
    relation = Relation("foo")
    assert_inconsistent(
        State(),
        _Event("foo_relation_changed", relation=relation),
        _CharmSpec(MyCharm, {"requires": {"foo": {"interface": "bar"}}}),
    )
    assert_consistent(
        State(relations={relation}),
        _Event("foo_relation_changed", relation=relation),
        _CharmSpec(MyCharm, {"requires": {"foo": {"interface": "bar"}}}),
    )


def test_action_not_in_meta_inconsistent():
    ctx = Context(MyCharm, meta={"name": "foo"}, actions={"foo": {}})
    assert_inconsistent(
        State(),
        ctx.on.action("foo", params={"bar": "baz"}),
        _CharmSpec(MyCharm, meta={}, actions={}),
    )


def test_action_meta_type_inconsistent():
    ctx = Context(MyCharm, meta={"name": "foo"}, actions={"foo": {}})
    assert_inconsistent(
        State(),
        ctx.on.action("foo", params={"bar": "baz"}),
        _CharmSpec(
            MyCharm, meta={}, actions={"foo": {"params": {"bar": {"type": "zabazaba"}}}}
        ),
    )

    assert_inconsistent(
        State(),
        ctx.on.action("foo", params={"bar": "baz"}),
        _CharmSpec(MyCharm, meta={}, actions={"foo": {"params": {"bar": {}}}}),
    )


def test_action_name():
    ctx = Context(MyCharm, meta={"name": "foo"}, actions={"foo": {}})

    assert_consistent(
        State(),
        ctx.on.action("foo", params={"bar": "baz"}),
        _CharmSpec(
            MyCharm, meta={}, actions={"foo": {"params": {"bar": {"type": "string"}}}}
        ),
    )
    assert_inconsistent(
        State(),
        _Event("box_action", action=ctx.on.action("foo", params={"bar": "baz"})),
        _CharmSpec(MyCharm, meta={}, actions={"foo": {}}),
    )


_ACTION_TYPE_CHECKS = [
    ("string", "baz", None),
    ("boolean", True, "baz"),
    ("integer", 42, 1.5),
    ("number", 28.8, "baz"),
    ("array", ["a", "b", "c"], 1.5),  # A string is an acceptable array.
    ("object", {"k": "v"}, "baz"),
]


@pytest.mark.parametrize("ptype,good,bad", _ACTION_TYPE_CHECKS)
def test_action_params_type(ptype, good, bad):
    ctx = Context(MyCharm, meta={"name": "foo"}, actions={"foo": {}})
    assert_consistent(
        State(),
        ctx.on.action("foo", params={"bar": good}),
        _CharmSpec(
            MyCharm, meta={}, actions={"foo": {"params": {"bar": {"type": ptype}}}}
        ),
    )
    if bad is not None:
        assert_inconsistent(
            State(),
            ctx.on.action("foo", params={"bar": bad}),
            _CharmSpec(
                MyCharm, meta={}, actions={"foo": {"params": {"bar": {"type": ptype}}}}
            ),
        )


def test_duplicate_relation_ids():
    assert_inconsistent(
        State(relations={Relation("foo", id=1), Relation("bar", id=1)}),
        _Event("start"),
        _CharmSpec(
            MyCharm,
            meta={
                "requires": {"foo": {"interface": "foo"}, "bar": {"interface": "bar"}}
            },
        ),
    )


def test_relation_without_endpoint():
    assert_inconsistent(
        State(relations={Relation("foo", id=1), Relation("bar", id=1)}),
        _Event("start"),
        _CharmSpec(MyCharm, meta={"name": "charlemagne"}),
    )

    assert_consistent(
        State(relations={Relation("foo", id=1), Relation("bar", id=2)}),
        _Event("start"),
        _CharmSpec(
            MyCharm,
            meta={
                "requires": {"foo": {"interface": "foo"}, "bar": {"interface": "bar"}}
            },
        ),
    )


def test_storage_event():
    storage = Storage("foo")
    assert_inconsistent(
        State(storages={storage}),
        _Event("foo-storage-attached"),
        _CharmSpec(MyCharm, meta={"name": "rupert"}),
    )
    assert_inconsistent(
        State(storages={storage}),
        _Event("foo-storage-attached"),
        _CharmSpec(
            MyCharm, meta={"name": "rupert", "storage": {"foo": {"type": "filesystem"}}}
        ),
    )


def test_storage_states():
    storage1 = Storage("foo", index=1)
    storage2 = Storage("foo", index=1)

    assert_inconsistent(
        State(storages={storage1, storage2}),
        _Event("start"),
        _CharmSpec(MyCharm, meta={"name": "everett"}),
    )
    assert_consistent(
        State(storages={storage1, dataclasses.replace(storage2, index=2)}),
        _Event("start"),
        _CharmSpec(
            MyCharm, meta={"name": "frank", "storage": {"foo": {"type": "filesystem"}}}
        ),
    )
    assert_consistent(
        State(storages={storage1, dataclasses.replace(storage2, name="marx")}),
        _Event("start"),
        _CharmSpec(
            MyCharm,
            meta={
                "name": "engels",
                "storage": {
                    "foo": {"type": "filesystem"},
                    "marx": {"type": "filesystem"},
                },
            },
        ),
    )


def test_storage_not_in_state():
    storage = Storage("foo")
    assert_inconsistent(
        State(),
        _Event("foo_storage_attached", storage=storage),
        _CharmSpec(
            MyCharm,
            meta={"name": "sam", "storage": {"foo": {"type": "filesystem"}}},
        ),
    )
    assert_consistent(
        State(storages=[storage]),
        _Event("foo_storage_attached", storage=storage),
        _CharmSpec(
            MyCharm,
            meta={"name": "sam", "storage": {"foo": {"type": "filesystem"}}},
        ),
    )


def test_resource_states():
    # happy path
    assert_consistent(
        State(resources={Resource(name="foo", path="/foo/bar.yaml")}),
        _Event("start"),
        _CharmSpec(
            MyCharm,
            meta={"name": "yamlman", "resources": {"foo": {"type": "oci-image"}}},
        ),
    )

    # no resources in state but some in meta: OK. Not realistic wrt juju but fine for testing
    assert_consistent(
        State(),
        _Event("start"),
        _CharmSpec(
            MyCharm,
            meta={"name": "yamlman", "resources": {"foo": {"type": "oci-image"}}},
        ),
    )

    # resource not defined in meta
    assert_inconsistent(
        State(resources={Resource(name="bar", path="/foo/bar.yaml")}),
        _Event("start"),
        _CharmSpec(
            MyCharm,
            meta={"name": "yamlman", "resources": {"foo": {"type": "oci-image"}}},
        ),
    )

    assert_inconsistent(
        State(resources={Resource(name="bar", path="/foo/bar.yaml")}),
        _Event("start"),
        _CharmSpec(
            MyCharm,
            meta={"name": "yamlman"},
        ),
    )


def test_networks_consistency():
    assert_inconsistent(
        State(networks={Network("foo")}),
        _Event("start"),
        _CharmSpec(
            MyCharm,
            meta={"name": "wonky"},
        ),
    )

    assert_inconsistent(
        State(networks={Network("foo")}),
        _Event("start"),
        _CharmSpec(
            MyCharm,
            meta={
                "name": "pinky",
                "extra-bindings": {"foo": {}},
                "requires": {"foo": {"interface": "bar"}},
            },
        ),
    )

    assert_consistent(
        State(networks={Network("foo")}),
        _Event("start"),
        _CharmSpec(
            MyCharm,
            meta={
                "name": "pinky",
                "extra-bindings": {"foo": {}},
                "requires": {"bar": {"interface": "bar"}},
            },
        ),
    )


def test_cloudspec_consistency():
    cloud_spec = CloudSpec(
        name="localhost",
        type="lxd",
        endpoint="https://127.0.0.1:8443",
        credential=CloudCredential(
            auth_type="clientcertificate",
            attributes={
                "client-cert": "foo",
                "client-key": "bar",
                "server-cert": "baz",
            },
        ),
    )

    assert_consistent(
        State(model=Model(name="lxd-model", type="lxd", cloud_spec=cloud_spec)),
        _Event("start"),
        _CharmSpec(
            MyCharm,
            meta={"name": "MyVMCharm"},
        ),
    )

    assert_inconsistent(
        State(model=Model(name="k8s-model", type="kubernetes", cloud_spec=cloud_spec)),
        _Event("start"),
        _CharmSpec(
            MyCharm,
            meta={"name": "MyK8sCharm"},
        ),
    )


def test_storedstate_consistency():
    assert_consistent(
        State(
            stored_states={
                StoredState(content={"foo": "bar"}),
                StoredState(name="my_stored_state", content={"foo": 1}),
                StoredState(owner_path="MyCharmLib", content={"foo": None}),
                StoredState(owner_path="OtherCharmLib", content={"foo": (1, 2, 3)}),
            }
        ),
        _Event("start"),
        _CharmSpec(
            MyCharm,
            meta={
                "name": "foo",
            },
        ),
    )
    assert_inconsistent(
        State(
            stored_states={StoredState(owner_path=None, content={"secret": Secret({})})}
        ),
        _Event("start"),
        _CharmSpec(
            MyCharm,
            meta={
                "name": "foo",
            },
        ),
    )



#!/usr/bin/env python3
# Copyright 2023 Canonical Ltd.
# See LICENSE file for licensing details.

"""Charm state-transition testing SDK for Ops charms.

Write tests that declaratively define the Juju state all at once, define the
Juju context against which to test the charm, and fire a single event on the
charm to execute its logic. The tests can then assert that the Juju state has
changed as expected.

These tests are 'state-transition' tests, a way to test isolated units of charm
functionality (how the state changes in reaction to events). They are not
necessarily tests of individual methods or functions (but might be, depending on
the charm's event observers); they are testing the 'contract' of the charm: given
a certain state, when a certain event happens, the charm should transition to a
certain (likely different) state. They do not test against a real Juju
controller and model, and focus on a single Juju unit, unlike integration tests.
For simplicity, we refer to them as 'unit' tests in the charm context.

Writing these tests should nudge you into thinking of a charm as a black-box
input->output function. The input is the union of an `Event` (why am I, charm,
being executed), a `State` (am I leader? what is my relation data? what is my
config?...) and the charm's execution `Context` (what relations can I have?
what containers can I have?...). The output is another `State`: the state after
the charm has had a chance to interact with the mocked Juju model and affect the
state.

.. image:: https://raw.githubusercontent.com/canonical/ops-scenario/main/resources/state-transition-model.png
   :alt: Transition diagram, with the input state and event on the left, the context including the charm in the centre, and the state out on the right

Writing unit tests for a charm, then, means verifying that:

- the charm does not raise uncaught exceptions while handling the event
- the output state (as compared with the input state) is as expected.

A test consists of three broad steps:

- **Arrange**:
    - declare the context
    - declare the input state
- **Act**:
    - select an event to fire
    - run the context (i.e. obtain the output state, given the input state and the event)
- **Assert**:
    - verify that the output state (as compared with the input state) is how you expect it to be
    - verify that the charm has seen a certain sequence of statuses, events, and `juju-log` calls
    - optionally, you can use a context manager to get a hold of the charm instance and run
      assertions on APIs and state internal to it.

The most basic scenario is one in which all is defaulted and barely any data is
available. The charm has no config, no relations, no leadership, and its
status is `unknown`. With that, we can write the simplest possible test:

.. code-block:: python

    def test_base():
        ctx = Context(MyCharm)
        out = ctx.run(ctx.on.start(), State())
        assert out.unit_status == UnknownStatus()
"""

from ops._private.harness import ActionFailed  # For backwards compatibility.

from .context import CharmEvents, Context, Manager
from .errors import StateValidationError  # For backwards compatibility.
from .state import (
    ActiveStatus,
    Address,
    AnyJson,
    BindAddress,
    BlockedStatus,
    CharmType,
    CheckInfo,
    CloudCredential,
    CloudSpec,
    Container,
    DeferredEvent,
    ErrorStatus,
    Exec,
    ICMPPort,
    JujuLogLine,
    MaintenanceStatus,
    Model,
    Mount,
    Network,
    Notice,
    PeerRelation,
    Port,
    RawDataBagContents,
    RawSecretRevisionContents,
    Relation,
    RelationBase,
    Resource,
    Secret,
    State,
    Storage,
    StoredState,
    SubordinateRelation,
    TCPPort,
    UDPPort,
    UnitID,
    UnknownStatus,
    WaitingStatus,
)

__all__ = [
    "ActionFailed",
    "ActiveStatus",
    "Address",
    "AnyJson",
    "BindAddress",
    "BlockedStatus",
    "CharmEvents",
    "CharmType",
    "CheckInfo",
    "CloudCredential",
    "CloudSpec",
    "Container",
    "Context",
    "DeferredEvent",
    "ErrorStatus",
    "Exec",
    "ICMPPort",
    "JujuLogLine",
    "MaintenanceStatus",
    "Manager",
    "Model",
    "Mount",
    "Network",
    "Notice",
    "PeerRelation",
    "Port",
    "RawDataBagContents",
    "RawSecretRevisionContents",
    "Relation",
    "RelationBase",
    "Resource",
    "Secret",
    "State",
    "StateValidationError",
    "Storage",
    "StoredState",
    "SubordinateRelation",
    "TCPPort",
    "UDPPort",
    "UnitID",
    "UnknownStatus",
    "WaitingStatus",
]



#!/usr/bin/env python3
# Copyright 2023 Canonical Ltd.
# See LICENSE file for licensing details.

"""Test framework runtime."""

import copy
import dataclasses
import os
import tempfile
import typing
from contextlib import contextmanager
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Dict,
    List,
    Optional,
    Type,
    TypeVar,
    Union,
)

import yaml
from ops import (
    CollectStatusEvent,
    pebble,
    CommitEvent,
    EventBase,
    Framework,
    Handle,
    NoTypeError,
    PreCommitEvent,
)
from ops.jujucontext import _JujuContext
from ops._private.harness import ActionFailed

from .errors import NoObserverError, UncaughtCharmError
from .logger import logger as scenario_logger
from .state import (
    PeerRelation,
    Relation,
    SubordinateRelation,
)

if TYPE_CHECKING:  # pragma: no cover
    from .context import Context
    from .state import CharmType, State, _CharmSpec, _Event

logger = scenario_logger.getChild("runtime")

RUNTIME_MODULE = Path(__file__).parent


class Runtime:
    """Charm runtime wrapper.

    This object bridges a local environment and a charm artifact.
    """

    def __init__(
        self,
        charm_spec: "_CharmSpec[CharmType]",
        charm_root: Optional[Union[str, Path]] = None,
        juju_version: str = "3.0.0",
        app_name: Optional[str] = None,
        unit_id: Optional[int] = 0,
    ):
        self._charm_spec = charm_spec
        self._juju_version = juju_version
        self._charm_root = charm_root

        app_name = app_name or self._charm_spec.meta.get("name")
        if not app_name:
            raise ValueError('invalid metadata: mandatory "name" field is missing.')

        self._app_name = app_name
        self._unit_id = unit_id

    def _get_event_env(self, state: "State", event: "_Event", charm_root: Path):
        """Build the simulated environment the operator framework expects."""
        env = {
            "JUJU_VERSION": self._juju_version,
            "JUJU_UNIT_NAME": f"{self._app_name}/{self._unit_id}",
            "_": "./dispatch",
            "JUJU_DISPATCH_PATH": f"hooks/{event.name}",
            "JUJU_MODEL_NAME": state.model.name,
            "JUJU_MODEL_UUID": state.model.uuid,
            "JUJU_CHARM_DIR": str(charm_root.absolute()),
        }

        if event._is_action_event and (action := event.action):
            env.update(
                {
                    "JUJU_ACTION_NAME": action.name.replace("_", "-"),
                    "JUJU_ACTION_UUID": action.id,
                },
            )

        if event._is_relation_event and (relation := event.relation):
            if isinstance(relation, PeerRelation):
                remote_app_name = self._app_name
            elif isinstance(relation, (Relation, SubordinateRelation)):
                remote_app_name = relation.remote_app_name
            else:
                raise ValueError(f"Unknown relation type: {relation}")
            env.update(
                {
                    "JUJU_RELATION": relation.endpoint,
                    "JUJU_RELATION_ID": str(relation.id),
                    "JUJU_REMOTE_APP": remote_app_name,
                },
            )

            remote_unit_id = event.relation_remote_unit_id

            # don't check truthiness because remote_unit_id could be 0
            if remote_unit_id is None and not event.name.endswith(
                ("_relation_created", "relation_broken"),
            ):
                remote_unit_ids = relation._remote_unit_ids

                if len(remote_unit_ids) == 1:
                    remote_unit_id = remote_unit_ids[0]
                    logger.info(
                        "there's only one remote unit, so we set JUJU_REMOTE_UNIT to it, "
                        "but you probably should be parametrizing the event with `remote_unit_id` "
                        "to be explicit.",
                    )
                elif len(remote_unit_ids) > 1:
                    remote_unit_id = remote_unit_ids[0]
                    logger.warning(
                        "remote unit ID unset, and multiple remote unit IDs are present; "
                        "We will pick the first one and hope for the best. You should be passing "
                        "`remote_unit_id` to the Event constructor.",
                    )
                else:
                    logger.warning(
                        "remote unit ID unset; no remote unit data present. "
                        "Is this a realistic scenario?",
                    )

            if remote_unit_id is not None:
                remote_unit = f"{remote_app_name}/{remote_unit_id}"
                env["JUJU_REMOTE_UNIT"] = remote_unit
                if event.name.endswith("_relation_departed"):
                    if event.relation_departed_unit_id:
                        env["JUJU_DEPARTING_UNIT"] = (
                            f"{remote_app_name}/{event.relation_departed_unit_id}"
                        )
                    else:
                        env["JUJU_DEPARTING_UNIT"] = remote_unit

        if container := event.container:
            env.update({"JUJU_WORKLOAD_NAME": container.name})

        if notice := event.notice:
            if hasattr(notice.type, "value"):
                notice_type = typing.cast(pebble.NoticeType, notice.type).value
            else:
                notice_type = str(notice.type)
            env.update(
                {
                    "JUJU_NOTICE_ID": notice.id,
                    "JUJU_NOTICE_TYPE": notice_type,
                    "JUJU_NOTICE_KEY": notice.key,
                },
            )

        if check_info := event.check_info:
            env["JUJU_PEBBLE_CHECK_NAME"] = check_info.name

        if storage := event.storage:
            env.update({"JUJU_STORAGE_ID": f"{storage.name}/{storage.index}"})

        if secret := event.secret:
            env.update(
                {
                    "JUJU_SECRET_ID": secret.id,
                    "JUJU_SECRET_LABEL": secret.label or "",
                },
            )
            # Don't check truthiness because revision could be 0.
            if event.secret_revision is not None:
                env["JUJU_SECRET_REVISION"] = str(event.secret_revision)

        return env

    @staticmethod
    def _wrap(charm_type: Type["CharmType"]) -> Type["CharmType"]:
        # dark sorcery to work around framework using class attrs to hold on to event sources
        # this should only be needed if we call play multiple times on the same runtime.
        class WrappedEvents(charm_type.on.__class__):
            """The charm's event sources, but wrapped."""

        WrappedEvents.__name__ = charm_type.on.__class__.__name__

        class WrappedCharm(charm_type):
            """The test charm's type, but with events wrapped."""

            on = WrappedEvents()

        WrappedCharm.__name__ = charm_type.__name__
        return typing.cast(Type["CharmType"], WrappedCharm)

    @contextmanager
    def _virtual_charm_root(self):
        # If we are using runtime on a real charm, we can make some assumptions about the
        # directory structure we are going to find.
        # If we're, say, dynamically defining charm types and doing tests on them, we'll have to
        # generate the metadata files ourselves. To be sure, we ALWAYS use a tempdir. Ground truth
        # is what the user passed via the CharmSpec
        spec = self._charm_spec

        if charm_virtual_root := self._charm_root:
            charm_virtual_root_is_custom = True
            virtual_charm_root = Path(charm_virtual_root)
        else:
            charm_virtual_root = tempfile.TemporaryDirectory()
            virtual_charm_root = Path(charm_virtual_root.name)
            charm_virtual_root_is_custom = False

        metadata_yaml = virtual_charm_root / "metadata.yaml"
        config_yaml = virtual_charm_root / "config.yaml"
        actions_yaml = virtual_charm_root / "actions.yaml"

        metadata_files_present: Dict[Path, Optional[str]] = {
            file: file.read_text()
            if charm_virtual_root_is_custom and file.exists()
            else None
            for file in (metadata_yaml, config_yaml, actions_yaml)
        }

        any_metadata_files_present_in_charm_virtual_root = any(
            v is not None for v in metadata_files_present.values()
        )

        if spec.is_autoloaded and charm_virtual_root_is_custom:
            # since the spec is autoloaded, in theory the metadata contents won't differ, so we can
            # overwrite away even if the custom vroot is the real charm root (the local repo).
            # Still, log it for clarity.
            if any_metadata_files_present_in_charm_virtual_root:
                logger.debug(
                    f"metadata files found in custom charm_root {charm_virtual_root}. "
                    f"The spec was autoloaded so the contents should be identical. "
                    f"Proceeding...",
                )

        elif (
            not spec.is_autoloaded and any_metadata_files_present_in_charm_virtual_root
        ):
            logger.warning(
                f"Some metadata files found in custom user-provided charm_root "
                f"{charm_virtual_root} while you have passed meta, config or actions to "
                f"Context.run(). "
                "Single source of truth are the arguments passed to Context.run(). "
                "charm_root metadata files will be overwritten for the "
                "duration of this test, and restored afterwards. "
                "To avoid this, clean any metadata files from the charm_root before calling run.",
            )

        metadata_yaml.write_text(yaml.safe_dump(spec.meta))
        config_yaml.write_text(yaml.safe_dump(spec.config or {}))
        actions_yaml.write_text(yaml.safe_dump(spec.actions or {}))

        yield virtual_charm_root

        if charm_virtual_root_is_custom:
            for file, previous_content in metadata_files_present.items():
                if previous_content is None:  # None == file did not exist before
                    file.unlink()
                else:
                    file.write_text(previous_content)

        else:
            # charm_virtual_root is a tempdir
            typing.cast(tempfile.TemporaryDirectory, charm_virtual_root).cleanup()  # type: ignore

    @contextmanager
    def _exec_ctx(self, ctx: "Context"):
        """python 3.8 compatibility shim"""
        with self._virtual_charm_root() as temporary_charm_root:
            with capture_events(
                include_deferred=ctx.capture_deferred_events,
                include_framework=ctx.capture_framework_events,
            ) as captured:
                yield (temporary_charm_root, captured)

    @contextmanager
    def exec(
        self,
        state: "State",
        event: "_Event",
        context: "Context",
    ):
        """Runs an event with this state as initial state on a charm.

        Returns the 'output state', that is, the state as mutated by the charm during the
        event handling.

        This will set the environment up and call ops.main().
        After that it's up to ops.
        """
        from ._consistency_checker import check_consistency  # avoid cycles

        check_consistency(state, event, self._charm_spec, self._juju_version)

        charm_type = self._charm_spec.charm_type
        logger.info(f"Preparing to fire {event.name} on {charm_type.__name__}")

        # we make a copy to avoid mutating the input state
        output_state = copy.deepcopy(state)

        logger.info(" - generating virtual charm root")
        with self._exec_ctx(context) as (temporary_charm_root, captured):
            logger.info(" - preparing env")
            env = self._get_event_env(
                state=state,
                event=event,
                charm_root=temporary_charm_root,
            )
            juju_context = _JujuContext.from_dict(env)
            # We need to set JUJU_VERSION in os.environ, because charms may use
            # `JujuVersion.from_environ()` to get the (simulated) Juju version.
            # For consistency, we put all the other ones there too, although we'd
            # like to change this in the future.
            previous_env = os.environ.copy()
            os.environ.update(env)

            logger.info(" - entering ops.main (mocked)")
            from ._ops_main_mock import Ops  # noqa: F811

            try:
                ops = Ops(
                    state=output_state,
                    event=event,
                    context=context,
                    charm_spec=dataclasses.replace(
                        self._charm_spec,
                        charm_type=self._wrap(charm_type),
                    ),
                    juju_context=juju_context,
                )

                yield ops

            except (NoObserverError, ActionFailed):
                raise  # propagate along
            except Exception as e:
                raise UncaughtCharmError(
                    f"Uncaught exception ({type(e)}) in operator/charm code: {e!r}",
                ) from e

            finally:
                for key in tuple(os.environ):
                    if key not in previous_env:
                        del os.environ[key]
                os.environ.update(previous_env)
                logger.info(" - exited ops.main")

        context.emitted_events.extend(captured)
        logger.info("event dispatched. done.")
        context._set_output_state(ops.state)


_T = TypeVar("_T", bound=EventBase)


@contextmanager
def capture_events(
    *types: Type[EventBase],
    include_framework: bool = False,
    include_deferred: bool = True,
):
    """Capture all events of type `*types` (using instance checks).

    Arguments exposed so that you can define your own fixtures if you want to.

    Example::
    >>> from ops import StartEvent
    >>> from scenario import Event, State
    >>> from charm import MyCustomEvent, MyCharm  # noqa
    >>>
    >>> def test_my_event():
    >>>     with capture_events(StartEvent, MyCustomEvent) as captured:
    >>>         trigger(State(), ("start", MyCharm, meta=MyCharm.META)
    >>>
    >>>     assert len(captured) == 2
    >>>     e1, e2 = captured
    >>>     assert isinstance(e2, MyCustomEvent)
    >>>     assert e2.custom_attr == 'foo'
    """
    allowed_types = types or (EventBase,)

    captured: List[EventBase] = []
    _real_emit = Framework._emit
    _real_reemit = Framework.reemit

    def _wrapped_emit(self: Framework, evt: EventBase):
        if not include_framework and isinstance(
            evt,
            (PreCommitEvent, CommitEvent, CollectStatusEvent),
        ):
            return _real_emit(self, evt)

        if isinstance(evt, allowed_types):
            # dump/undump the event to ensure any custom attributes are (re)set by restore()
            evt.restore(evt.snapshot())
            captured.append(evt)

        return _real_emit(self, evt)

    def _wrapped_reemit(self: Framework):
        # Framework calls reemit() before emitting the main juju event. We intercept that call
        # and capture all events in storage.

        if not include_deferred:
            return _real_reemit(self)

        # load all notices from storage as events.
        for event_path, _, _ in self._storage.notices():
            event_handle = Handle.from_path(event_path)
            try:
                event = self.load_snapshot(event_handle)
            except NoTypeError:
                continue
            event = typing.cast(EventBase, event)
            event.deferred = False
            self._forget(event)  # prevent tracking conflicts

            if not include_framework and isinstance(
                event,
                (PreCommitEvent, CommitEvent),
            ):
                continue

            if isinstance(event, allowed_types):
                captured.append(event)

        return _real_reemit(self)

    Framework._emit = _wrapped_emit  # type: ignore
    Framework.reemit = _wrapped_reemit

    yield captured

    Framework._emit = _real_emit
    Framework.reemit = _real_reemit



import copy
import typing

import ops
import pytest

import scenario

META = {
    "name": "context-charm",
    "containers": {
        "bar": {},
    },
    "requires": {
        "baz": {
            "interface": "charmlink",
        }
    },
    "storage": {
        "foo": {
            "type": "filesystem",
        }
    },
}
ACTIONS = {
    "act": {
        "params": {
            "param": {
                "description": "some parameter",
                "type": "string",
                "default": "",
            }
        }
    },
}


class ContextCharm(ops.CharmBase):
    def __init__(self, framework: ops.Framework):
        super().__init__(framework)
        self.observed: typing.List[ops.EventBase] = []
        for event in self.on.events().values():
            framework.observe(event, self._on_event)

    def _on_event(self, event: ops.EventBase):
        self.observed.append(event)


@pytest.mark.parametrize(
    "event_name, event_kind",
    [
        ("install", ops.InstallEvent),
        ("start", ops.StartEvent),
        ("stop", ops.StopEvent),
        ("remove", ops.RemoveEvent),
        ("update_status", ops.UpdateStatusEvent),
        ("config_changed", ops.ConfigChangedEvent),
        ("upgrade_charm", ops.UpgradeCharmEvent),
        ("leader_elected", ops.LeaderElectedEvent),
    ],
)
def test_simple_events(event_name: str, event_kind: typing.Type[ops.EventBase]):
    ctx = scenario.Context(ContextCharm, meta=META, actions=ACTIONS)
    # These look like:
    #   ctx.run(ctx.on.install(), state)
    with ctx(getattr(ctx.on, event_name)(), scenario.State()) as mgr:
        mgr.run()
        assert len(mgr.charm.observed) == 2
        assert isinstance(mgr.charm.observed[1], ops.CollectStatusEvent)
        assert isinstance(mgr.charm.observed[0], event_kind)


@pytest.mark.parametrize(
    "event_name, event_kind",
    [
        ("pre_series_upgrade", ops.PreSeriesUpgradeEvent),
        ("post_series_upgrade", ops.PostSeriesUpgradeEvent),
    ],
)
def test_simple_deprecated_events(event_name, event_kind):
    ctx = scenario.Context(ContextCharm, meta=META, actions=ACTIONS)
    # These look like:
    #   ctx.run(ctx.on.pre_series_upgrade(), state)
    with pytest.warns(DeprecationWarning):
        with ctx(getattr(ctx.on, event_name)(), scenario.State()) as mgr:
            mgr.run()
            assert len(mgr.charm.observed) == 2
            assert isinstance(mgr.charm.observed[1], ops.CollectStatusEvent)
            assert isinstance(mgr.charm.observed[0], event_kind)


@pytest.mark.parametrize("as_kwarg", [True, False])
@pytest.mark.parametrize(
    "event_name,event_kind,owner",
    [
        ("secret_changed", ops.SecretChangedEvent, None),
        ("secret_rotate", ops.SecretRotateEvent, "app"),
    ],
)
def test_simple_secret_events(as_kwarg, event_name, event_kind, owner):
    ctx = scenario.Context(ContextCharm, meta=META, actions=ACTIONS)
    secret = scenario.Secret({"password": "xxxx"}, owner=owner)
    state_in = scenario.State(secrets={secret})
    # These look like:
    #   ctx.run(ctx.on.secret_changed(secret=secret), state)
    # The secret must always be passed because the same event name is used for
    # all secrets.
    if as_kwarg:
        args = ()
        kwargs = {"secret": secret}
    else:
        args = (secret,)
        kwargs = {}
    with ctx(getattr(ctx.on, event_name)(*args, **kwargs), state_in) as mgr:
        mgr.run()
        assert len(mgr.charm.observed) == 2
        assert isinstance(mgr.charm.observed[1], ops.CollectStatusEvent)
        mgr = mgr.charm.observed[0]
        assert isinstance(mgr, event_kind)
        assert mgr.secret.id == secret.id


@pytest.mark.parametrize(
    "event_name, event_kind",
    [
        ("secret_expired", ops.SecretExpiredEvent),
        ("secret_remove", ops.SecretRemoveEvent),
    ],
)
def test_revision_secret_events(event_name, event_kind):
    ctx = scenario.Context(ContextCharm, meta=META, actions=ACTIONS)
    secret = scenario.Secret(
        tracked_content={"password": "yyyy"},
        latest_content={"password": "xxxx"},
        owner="app",
    )
    state_in = scenario.State(secrets={secret})
    # These look like:
    #   ctx.run(ctx.on.secret_expired(secret=secret, revision=revision), state)
    # The secret and revision must always be passed because the same event name
    # is used for all secrets.
    with ctx(getattr(ctx.on, event_name)(secret, revision=42), state_in) as mgr:
        mgr.run()
        assert len(mgr.charm.observed) == 2
        assert isinstance(mgr.charm.observed[1], ops.CollectStatusEvent)
        mgr = mgr.charm.observed[0]
        assert isinstance(mgr, event_kind)
        assert mgr.secret.id == secret.id
        assert mgr.revision == 42


@pytest.mark.parametrize("event_name", ["secret_expired", "secret_remove"])
def test_revision_secret_events_as_positional_arg(event_name):
    ctx = scenario.Context(ContextCharm, meta=META, actions=ACTIONS)
    secret = scenario.Secret(
        tracked_content={"password": "yyyy"},
        latest_content={"password": "xxxx"},
        owner=None,
    )
    state_in = scenario.State(secrets={secret})
    with pytest.raises(TypeError):
        ctx.run(getattr(ctx.on, event_name)(secret, 42), state_in)


@pytest.mark.parametrize(
    "event_name, event_kind",
    [
        ("storage_attached", ops.StorageAttachedEvent),
        ("storage_detaching", ops.StorageDetachingEvent),
    ],
)
def test_storage_events(event_name, event_kind):
    ctx = scenario.Context(ContextCharm, meta=META, actions=ACTIONS)
    storage = scenario.Storage("foo")
    state_in = scenario.State(storages=[storage])
    # These look like:
    #   ctx.run(ctx.on.storage_attached(storage), state)
    with ctx(getattr(ctx.on, event_name)(storage), state_in) as mgr:
        mgr.run()
        assert len(mgr.charm.observed) == 2
        assert isinstance(mgr.charm.observed[1], ops.CollectStatusEvent)
        mgr = mgr.charm.observed[0]
        assert isinstance(mgr, event_kind)
        assert mgr.storage.name == storage.name
        assert mgr.storage.index == storage.index


def test_action_event_no_params():
    ctx = scenario.Context(ContextCharm, meta=META, actions=ACTIONS)
    # These look like:
    #   ctx.run(ctx.on.action(action_name), state)
    with ctx(ctx.on.action("act"), scenario.State()) as mgr:
        mgr.run()
        assert len(mgr.charm.observed) == 2
        assert isinstance(mgr.charm.observed[1], ops.CollectStatusEvent)
        mgr = mgr.charm.observed[0]
        assert isinstance(mgr, ops.ActionEvent)


def test_action_event_with_params():
    ctx = scenario.Context(ContextCharm, meta=META, actions=ACTIONS)
    # These look like:
    #   ctx.run(ctx.on.action(action=action), state)
    # So that any parameters can be included and the ID can be customised.
    call_event = ctx.on.action("act", params={"param": "hello"})
    with ctx(call_event, scenario.State()) as mgr:
        mgr.run()
        assert len(mgr.charm.observed) == 2
        assert isinstance(mgr.charm.observed[1], ops.CollectStatusEvent)
        mgr = mgr.charm.observed[0]
        assert isinstance(mgr, ops.ActionEvent)
        assert mgr.id == call_event.action.id
        assert mgr.params["param"] == call_event.action.params["param"]


def test_pebble_ready_event():
    ctx = scenario.Context(ContextCharm, meta=META, actions=ACTIONS)
    container = scenario.Container("bar", can_connect=True)
    state_in = scenario.State(containers=[container])
    # These look like:
    #   ctx.run(ctx.on.pebble_ready(container), state)
    with ctx(ctx.on.pebble_ready(container), state_in) as mgr:
        mgr.run()
        assert len(mgr.charm.observed) == 2
        assert isinstance(mgr.charm.observed[1], ops.CollectStatusEvent)
        mgr = mgr.charm.observed[0]
        assert isinstance(mgr, ops.PebbleReadyEvent)
        assert mgr.workload.name == container.name


@pytest.mark.parametrize("as_kwarg", [True, False])
@pytest.mark.parametrize(
    "event_name, event_kind",
    [
        ("relation_created", ops.RelationCreatedEvent),
        ("relation_broken", ops.RelationBrokenEvent),
    ],
)
def test_relation_app_events(as_kwarg, event_name, event_kind):
    ctx = scenario.Context(ContextCharm, meta=META, actions=ACTIONS)
    relation = scenario.Relation("baz")
    state_in = scenario.State(relations=[relation])
    # These look like:
    #   ctx.run(ctx.on.relation_created(relation), state)
    if as_kwarg:
        args = ()
        kwargs = {"relation": relation}
    else:
        args = (relation,)
        kwargs = {}
    with ctx(getattr(ctx.on, event_name)(*args, **kwargs), state_in) as mgr:
        mgr.run()
        assert len(mgr.charm.observed) == 2
        assert isinstance(mgr.charm.observed[1], ops.CollectStatusEvent)
        mgr = mgr.charm.observed[0]
        assert isinstance(mgr, event_kind)
        assert mgr.relation.id == relation.id
        assert mgr.app.name == relation.remote_app_name
        assert mgr.unit is None


def test_relation_complex_name():
    meta = copy.deepcopy(META)
    meta["requires"]["foo-bar-baz"] = {"interface": "another-one"}
    ctx = scenario.Context(ContextCharm, meta=meta, actions=ACTIONS)
    relation = scenario.Relation("foo-bar-baz")
    state_in = scenario.State(relations=[relation])
    with ctx(ctx.on.relation_created(relation), state_in) as mgr:
        mgr.run()
        assert len(mgr.charm.observed) == 2
        mgr = mgr.charm.observed[0]
        assert isinstance(mgr, ops.RelationCreatedEvent)
        assert mgr.relation.id == relation.id
        assert mgr.app.name == relation.remote_app_name
        assert mgr.unit is None


@pytest.mark.parametrize("event_name", ["relation_created", "relation_broken"])
def test_relation_events_as_positional_arg(event_name):
    ctx = scenario.Context(ContextCharm, meta=META, actions=ACTIONS)
    relation = scenario.Relation("baz")
    state_in = scenario.State(relations=[relation])
    with pytest.raises(TypeError):
        ctx.run(getattr(ctx.on, event_name)(relation, 0), state_in)


@pytest.mark.parametrize(
    "event_name, event_kind",
    [
        ("relation_joined", ops.RelationJoinedEvent),
        ("relation_changed", ops.RelationChangedEvent),
    ],
)
def test_relation_unit_events_default_unit(event_name, event_kind):
    ctx = scenario.Context(ContextCharm, meta=META, actions=ACTIONS)
    relation = scenario.Relation("baz", remote_units_data={1: {"x": "y"}})
    state_in = scenario.State(relations=[relation])
    # These look like:
    #   ctx.run(ctx.on.baz_relation_changed, state)
    # The unit is chosen automatically.
    with ctx(getattr(ctx.on, event_name)(relation), state_in) as mgr:
        mgr.run()
        assert len(mgr.charm.observed) == 2
        assert isinstance(mgr.charm.observed[1], ops.CollectStatusEvent)
        mgr = mgr.charm.observed[0]
        assert isinstance(mgr, event_kind)
        assert mgr.relation.id == relation.id
        assert mgr.app.name == relation.remote_app_name
        assert mgr.unit.name == "remote/1"


@pytest.mark.parametrize(
    "event_name, event_kind",
    [
        ("relation_joined", ops.RelationJoinedEvent),
        ("relation_changed", ops.RelationChangedEvent),
    ],
)
def test_relation_unit_events(event_name, event_kind):
    ctx = scenario.Context(ContextCharm, meta=META, actions=ACTIONS)
    relation = scenario.Relation(
        "baz", remote_units_data={1: {"x": "y"}, 2: {"x": "z"}}
    )
    state_in = scenario.State(relations=[relation])
    # These look like:
    #   ctx.run(ctx.on.baz_relation_changed(unit=unit_ordinal), state)
    with ctx(getattr(ctx.on, event_name)(relation, remote_unit=2), state_in) as mgr:
        mgr.run()
        assert len(mgr.charm.observed) == 2
        assert isinstance(mgr.charm.observed[1], ops.CollectStatusEvent)
        mgr = mgr.charm.observed[0]
        assert isinstance(mgr, event_kind)
        assert mgr.relation.id == relation.id
        assert mgr.app.name == relation.remote_app_name
        assert mgr.unit.name == "remote/2"


def test_relation_departed_event():
    ctx = scenario.Context(ContextCharm, meta=META, actions=ACTIONS)
    relation = scenario.Relation("baz")
    state_in = scenario.State(relations=[relation])
    # These look like:
    #   ctx.run(ctx.on.baz_relation_departed(unit=unit_ordinal, departing_unit=unit_ordinal), state)
    with ctx(
        ctx.on.relation_departed(relation, remote_unit=2, departing_unit=1), state_in
    ) as mgr:
        mgr.run()
        assert len(mgr.charm.observed) == 2
        assert isinstance(mgr.charm.observed[1], ops.CollectStatusEvent)
        mgr = mgr.charm.observed[0]
        assert isinstance(mgr, ops.RelationDepartedEvent)
        assert mgr.relation.id == relation.id
        assert mgr.app.name == relation.remote_app_name
        assert mgr.unit.name == "remote/2"
        assert mgr.departing_unit.name == "remote/1"



import dataclasses

import pytest
from ops.charm import CharmBase
from ops.framework import Framework
from ops.model import ActiveStatus, BlockedStatus

from scenario.state import Relation, State
from ..helpers import jsonpatch_delta, trigger


@pytest.fixture(scope="function")
def mycharm():
    class MyCharm(CharmBase):
        _call = None
        called = False

        def __init__(self, framework: Framework):
            super().__init__(framework)

            for evt in self.on.events().values():
                self.framework.observe(evt, self._on_event)

        def _on_event(self, event):
            if MyCharm._call:
                MyCharm.called = True
                MyCharm._call(self, event)

    return MyCharm


def test_charm_heals_on_start(mycharm):
    def pre_event(charm):
        pre_event._called = True
        assert charm.unit.status == BlockedStatus("foo")
        assert not charm.called

    def call(charm, _):
        if charm.unit.status.message == "foo":
            charm.unit.status = ActiveStatus("yabadoodle")

    def post_event(charm):
        post_event._called = True
        assert charm.unit.status == ActiveStatus("yabadoodle")
        assert charm.called

    mycharm._call = call

    initial_state = State(
        config={"foo": "bar"}, leader=True, unit_status=BlockedStatus("foo")
    )

    out = trigger(
        initial_state,
        charm_type=mycharm,
        meta={"name": "foo"},
        config={"options": {"foo": {"type": "string"}}},
        event="start",
        post_event=post_event,
        pre_event=pre_event,
    )

    assert out.unit_status == ActiveStatus("yabadoodle")

    out_purged = dataclasses.replace(out, stored_states=initial_state.stored_states)
    assert jsonpatch_delta(out_purged, initial_state) == [
        {
            "op": "replace",
            "path": "/unit_status/message",
            "value": "yabadoodle",
        },
        {
            "op": "replace",
            "path": "/unit_status/name",
            "value": "active",
        },
    ]


def test_relation_data_access(mycharm):
    mycharm._call = lambda *_: True

    def check_relation_data(charm):
        foo_relations = charm.model.relations["relation_test"]
        assert len(foo_relations) == 1
        foo_rel = foo_relations[0]
        assert len(foo_rel.units) == 2

        remote_units_data = {}
        for remote_unit in foo_rel.units:
            remote_units_data[remote_unit.name] = dict(foo_rel.data[remote_unit])

        remote_app_data = foo_rel.data[foo_rel.app]

        assert remote_units_data == {
            "karlos/0": {"foo": "bar"},
            "karlos/1": {"baz": "qux"},
        }

        assert remote_app_data == {"yaba": "doodle"}

    state_in = State(
        relations={
            Relation(
                endpoint="relation_test",
                interface="azdrubales",
                id=1,
                remote_app_name="karlos",
                remote_app_data={"yaba": "doodle"},
                remote_units_data={0: {"foo": "bar"}, 1: {"baz": "qux"}},
            )
        }
    )
    trigger(
        state_in,
        charm_type=mycharm,
        meta={
            "name": "foo",
            "requires": {"relation_test": {"interface": "azdrubales"}},
        },
        event="update_status",
        post_event=check_relation_data,
    )



import ops
import pytest
from ops import CharmBase

from scenario import Context
from scenario.state import State, _CharmSpec, _Event, _EventType


@pytest.mark.parametrize(
    "evt, expected_type",
    (
        ("foo_relation_changed", _EventType.RELATION),
        ("foo_relation_created", _EventType.RELATION),
        ("foo_bar_baz_relation_created", _EventType.RELATION),
        ("foo_storage_attached", _EventType.STORAGE),
        ("foo_storage_detaching", _EventType.STORAGE),
        ("foo_bar_baz_storage_detaching", _EventType.STORAGE),
        ("foo_pebble_ready", _EventType.WORKLOAD),
        ("foo_bar_baz_pebble_ready", _EventType.WORKLOAD),
        ("foo_pebble_custom_notice", _EventType.WORKLOAD),
        ("foo_bar_baz_pebble_custom_notice", _EventType.WORKLOAD),
        ("secret_remove", _EventType.SECRET),
        ("pre_commit", _EventType.FRAMEWORK),
        ("commit", _EventType.FRAMEWORK),
        ("collect_unit_status", _EventType.FRAMEWORK),
        ("collect_app_status", _EventType.FRAMEWORK),
        ("foo", _EventType.CUSTOM),
        ("kaboozle_bar_baz", _EventType.CUSTOM),
    ),
)
def test_event_type(evt, expected_type):
    event = _Event(evt)
    assert event._path.type is expected_type

    assert event._is_relation_event is (expected_type is _EventType.RELATION)
    assert event._is_storage_event is (expected_type is _EventType.STORAGE)
    assert event._is_workload_event is (expected_type is _EventType.WORKLOAD)
    assert event._is_secret_event is (expected_type is _EventType.SECRET)
    assert event._is_action_event is (expected_type is _EventType.ACTION)

    class MyCharm(CharmBase):
        pass

    spec = _CharmSpec(
        MyCharm,
        meta={
            "requires": {
                "foo": {"interface": "bar"},
                "foo_bar_baz": {"interface": "bar"},
            },
            "storage": {
                "foo": {"type": "filesystem"},
                "foo_bar_baz": {"type": "filesystem"},
            },
            "containers": {"foo": {}, "foo_bar_baz": {}},
        },
    )
    assert event._is_builtin_event(spec) is (expected_type is not _EventType.CUSTOM)


def test_emitted_framework():
    class MyCharm(CharmBase):
        META = {"name": "joop"}

    ctx = Context(MyCharm, meta=MyCharm.META, capture_framework_events=True)
    ctx.run(ctx.on.update_status(), State())
    assert len(ctx.emitted_events) == 4
    assert list(map(type, ctx.emitted_events)) == [
        ops.UpdateStatusEvent,
        ops.CollectStatusEvent,
        ops.PreCommitEvent,
        ops.CommitEvent,
    ]


def test_emitted_deferred():
    class MyCharm(CharmBase):
        META = {"name": "joop"}

        def _foo(self, e):
            pass

    ctx = Context(
        MyCharm,
        meta=MyCharm.META,
        capture_deferred_events=True,
        capture_framework_events=True,
    )
    ctx.run(
        ctx.on.start(), State(deferred=[_Event("update-status").deferred(MyCharm._foo)])
    )

    assert len(ctx.emitted_events) == 5
    assert [e.handle.kind for e in ctx.emitted_events] == [
        "update_status",
        "start",
        "collect_unit_status",
        "pre_commit",
        "commit",
    ]
